{% if (location is defined) and location %}
no snmp-server location
{% endif %}
{% if (contact is defined) and contact %}
no snmp-server contact
{% endif %}
{% if (chassis_id is defined) and chassis_id %}
no snmp-server chassis-id
{% endif %}
{% if (community is defined) and community %}
{% for comm_name, comm_details in community.iteritems() %}
no snmp-server community {{comm_name}}
{% endfor %}
{% endif %}
