{%- for user_name, user_details in users.items() %}
{%- if user_details -%}
{%- if user_details.get('sshkeys') %}
{%- for sshkey in user_details.sshkeys %}
no username {{user_name}} sshkey {{ sshkey }}
{%- endfor %}
{%- endif %}
{%- if user_details.get('password') %}
no username {{user_name}} password
{%- endif %}
{%- if user_details.get('level') %}
no username {{user_name}} role priv-{{user_details.level}}
{%- endif %}
{%- endfor %}
{%- else -%}
{%- endif -%}
