#!/usr/bin/env python
# coding=utf-8
# Copyright 2015 Netflix. All rights reserved.
# Copyright 2016 BigWaveIT. All rights reserved.
#
# The contents of this file are licensed under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

"""
A module to interact with Cisco devices running IOS-XR.

Forked from https://github.com/fooelisa/pyiosxr 2019-10-27

Contributors fooelisa, mirceaulinic, ktbyers et all
"""

from napalm.pyIOSXR.iosxr import IOSXR

__all__ = ("IOSXR",)
