services {
  rpm {
    {% for probe_name, probe_test in probes.iteritems() %}
    probe {{probe_name}} {
      {% for test_name, test_details in probe_test.iteritems() %}
        delete: test {{test_name}};
      {% endfor %}
    }
    {% endfor %}
  }
}
